#########################################################################################
#########
############################################
###############################################################################################
##############################
######################################################################################################

#########
#############################################################
#############################
############################################################

import arcpy,os,traceback

try:
    arcpy.CheckOutExtension("spatial")
    #arcpy.env.cartographicCoordinateSystem = "PROJCS['USA_Contiguous_Albers_Equal_Area_Conic_USGS_version',GEOGCS['GCS_North_American_1983',DATUM['D_North_American_1983',SPHEROID['GRS_1980',6378137.0,298.257222101]],PRIMEM['Greenwich',0.0],UNIT['Degree',0.0174532925199433]],PROJECTION['Albers'],PARAMETER['False_Easting',0.0],PARAMETER['False_Northing',0.0],PARAMETER['Central_Meridian',-96.0],PARAMETER['Standard_Parallel_1',29.5],PARAMETER['Standard_Parallel_2',45.5],PARAMETER['Latitude_Of_Origin',23.0],UNIT['Meter',1.0]]"

    inGrid=arcpy.GetParameterAsText(0)
    ruGrid=arcpy.GetParameterAsText(1)
    ruFC=arcpy.GetParameterAsText(2)
    outFC=arcpy.GetParameterAsText(3)

    cellSize=float(arcpy.GetRasterProperties_management(ruGrid,"CELLSIZEX").getOutput(0))
    arcpy.env.cellSize=cellSize

    #scratch workspace must be set to a folder NOT a geodatabase
    if not arcpy.env.scratchWorkspace:
        arcpy.env.scratchWorkspace="c:/temp"
    elif arcpy.env.scratchWorkspace.count(".gdb"):
        arcpy.env.scratchWorkspace=os.path.join(os.path.split(arcpy.env.scratchWorkspace)[0],"temp")
    if not os.path.exists(arcpy.env.scratchWorkspace):
        os.makedirs(arcpy.env.scratchWorkspace)

    arcpy.AddMessage("Scratch Workspace: %s"%(arcpy.env.scratchWorkspace))
    tmpGrid=os.path.normpath("%s/tmpMultGd"%(arcpy.env.scratchWorkspace))

    for entry in [outFC,tmpGrid]:
        if arcpy.Exists(entry):
            arcpy.Delete_management(entry)  

    #Multiply grids
    arcpy.AddMessage("Multiplying grids...")
    arcpy.gp.Times_sa(inGrid,ruGrid,tmpGrid)
    #attribute table does not get built if there are too many reporting units
    arcpy.BuildRasterAttributeTable_management(tmpGrid)

    results=dict()
    rows = arcpy.SearchCursor(tmpGrid)
    for row in rows:
        #calculate area
        results[row.VALUE]=cellSize*cellSize*float(row.COUNT)
    del row,rows

    arcpy.AddMessage("Adding results to output...")
    arcpy.CopyFeatures_management(ruFC,outFC)
    arcpy.AddField_management(outFC, "DENSITY", "DOUBLE", 10, 4, "","", "NULLABLE")
    rows=arcpy.UpdateCursor(outFC)
    for row in rows:
        rowID=row.getValue("OBJECTID")
        value=0
        area=row.shape.area 
        if (results.has_key(rowID)) and area>0:
            value=min([100.0,100.0 * float(results[rowID])/area])
        row.setValue("DENSITY",value)
        rows.updateRow(row)
    del row,rows

    arcpy.Delete_management(tmpGrid)         
    arcpy.SetParameter(3,outFC)
    arcpy.SetParameter(4,"DENSITY")

except:
    arcpy.AddError(traceback.format_exc())
    print traceback.format_exc()
    raise Exception("Error occurred during processing")



