#################### copy_field_to_table.py ######################
# Copies field from one table to another table, with a new name, joining the two tables on OBJECTID

# Inputs:
# Input Table to copy field from (must match on OBJECTID)
# Input field name
# Output Table
# Output Field Name  (will destroy and recreate if already exists)
#
#Outputs:
# Output table (so it can be consumed by later tools)
# Output field name
############################################################

import arcpy,os,traceback

def RaiseException(errStr):
    try:
        raise Exception(errStr)
    except Exception as inst:
        print inst
    raise

try:
    # Get the inputs
    inTable = arcpy.GetParameterAsText(0)
    inField = arcpy.GetParameterAsText(1)
    outTable = arcpy.GetParameterAsText(2)
    outField = arcpy.GetParameterAsText(3)

    inTableRoot=os.path.split(inTable)[1]
    if outField=="DEFAULT":
        outField="%s_%s"%(inTableRoot,inField)

    fieldType,precision,scale="DOUBLE",10,4
    for field in arcpy.ListFields(inTable):
        if field.name==inField:
            fieldType,precision,scale=field.type,field.precision,field.scale
            if fieldType=="Single":
                fieldType="DOUBLE"
            
    for field in arcpy.ListFields(outTable):
        if field.name==outField:
            arcpy.AddWarning("Output Error: Field %s already exists in %s"%(outField,outTable))
            arcpy.DeleteField_management(outTable,[outField])
            
    arcpy.AddMessage("Creating field: %s"%(outField))
    arcpy.AddField_management(outTable,outField,fieldType,precision,scale,"","", "NULLABLE")
                              
    #add temporary join to intable and copy across the field we want into out field
    outTableView="outTableView"
    if arcpy.Exists(outTableView):
        arcpy.Delete_management(outTableView)
    arcpy.MakeTableView_management (outTable, outTableView)
    arcpy.AddJoin_management (outTableView, "OBJECTID", inTable, "OBJECTID")
    arcpy.CalculateField_management(outTableView,outField,"!%s.%s!"%(inTableRoot,inField),"PYTHON")        
    #join and table view automatically discarded on exit

    arcpy.SetParameter(4,outTable)
    arcpy.SetParameter(5,outField)

except:
    arcpy.AddError(traceback.format_exc())
    RaiseException("Error occurred during processing")



