from __future__ import division
import arcpy
import random
import time
import csv
import pickle
import re
import os
from arcpy import env


########################################################################
# Globals for control, etc
DEBUG = False
MODELBUILDER = True

MinForFuzzyLimit = -9999
MaxForFuzzyLimit = 9999
NullFuzzyValue = 0
########################################################################

def OutMsg(msg):
    if MODELBUILDER:
        arcpy.AddMessage('\t%s'%(msg))
    else:
        print(msg);
# def OutMsg(msg):

def ProfileTime(outStr):
    OutMsg("%7.3f %s"%(time.clock(),outStr))
# def ProfileTime(outStr):

def RaiseException(errStr):
    try:
        raise Exception(errStr)
    except Exception as inst:
        OutMsg(errStr)
    raise
# end def RaiseException(errStr):

def isFloat(s):
    try:
        float(s)
        return True
    except ValueError:
        return False
    except TypeError:
        return False
# def isFloat(s):

class EEMSTbl:
    tblDict = {}
    ndxDict = {}
    tblSz = -1

    def AddFld(self,fldNm,lst):
        if self.tblSz == -1:
            self.tblSz = len(lst)
        else:
            if len(lst) != self.tblSz:
                RaiseException("Error: Table size of %s does not match!"%(fldNm))

        self.tblDict[fldNm] = lst
    # def AddFld()

    def WriteFldsToCSV(self,fileName):
        outFile = open(fileName,'w')
        line = ','.join(self.tblDict.keys())

        outFile.write(line)
        outFile.write('\n')

        for i in range(self.tblSz):
            outTokens = []
            for key in self.tblDict.keys():
                outTokens.append(str(self.tblDict[key][i]))
                line = line = ','.join(outTokens)
            # for key in self.tblDict.keys():

            outFile.write(line)
            outFile.write('\n')

        # for i in range (self.tblSz):

        outFile.close()
    # def WriteFldsToCSV(self,fileName):

    def ReadFldsFromCSV(self,fileName):
        inFile = open(fileName,'r')
        line = inFile.readline()
        line = line.rstrip('\r\f\n')
        
        line = re.sub('"','',line)

        fldNms = line.split(',')

        for fldNm in fldNms:
            self.tblDict[fldNm] = []

        line = inFile.readline()

        while line != '':
            line = re.sub('"','',line)
            inTokens = line.split(',')
            i = 0
            for fldNm in fldNms:
                try:
                    self.tblDict[fldNm].append(float(inTokens[i]))
                except ValueError:
                    self.tblDict[fldNm].append(None)
                i = i+1
            # for fldNm in fldNms:
            line = inFile.readline()
        # while line != '':
            
        inFile.close()

        if self.tblSz == -1:
            self.tblSz = len(self.tblDict[fldNm])
        else:
            if self.tblSz != len(self.tblDict[fldNm]):
                RaiseException('ERROR! csv file %s (size %d) mismatched table size %d.'
                               %(fileName, len(self.tblDict[fldNm]), self.tblSz))
        # if tblSz == -1:

    # def ReadFldsFromCSV(self,fileName):

    def FldExists(self,fldNm):
        return fldNm in self.tblDict.keys()
    # def FldExists(self,fldNm):

    def NdxExists(self,fldNm):
        return fldNm in self.ndxDict.keys()
    # def NdxExists(self,fldNm):

    def ConfirmFldsExist(self,fldNms):
        for fldNm in fldNms:
            self.ConfirmFldExists(fldNm)
    # def ConfirmFldsExist(*fldNms):

    def ConfirmFldExists(self,fldNm):
        if fldNm not in self.tblDict.keys():
            RaiseException("Error: Field %s does not exist!"%(fldNm))
    # def ConfirmFldExists(fldNms):

    def ConfirmNdxsExist(self,ndxNms):
        for ndxNm in ndxNms:
            self.ConfirmNdxExists(ndxNm)
    # def ConfirmNdxExist(*fldNms):

    def ConfirmNdxExists(self,ndxNm):
        if ndxNm not in self.ndxDict.keys():
            RaiseException("Error: Index %s does not exist!"%(ndxNm))
    # def ConfirmNdxExist(fldNms):

    def IndexOnFld(self,fldNm):
        self.ndxDict[fldNm] = {}
        for i in range(self.tblSz):
            # Value in table is key, row number is index value
            ndxKey = self.tblDict[fldNm][i] 
            if self.ndxDict[fldNm].has_key(ndxKey):
                RaiseException("ERROR: Cannot create index on %s, non unique value."%(fldNm))
            else:
                self.ndxDict[fldNm][ndxKey] = i
    # def IndexOnFld(self,fldNm):

    def GetIndexedFldValue(self,ndxNm,ndxVal,fldNm):
        self.ConfirmNdxExists(ndxNm)
        self.ConfirmFldExists(fldNm)
        return self.tblDict[fldNm][self.ndxDict[ndxNm][ndxVal]]
    # def GetIndexedFldValue(self,ndxNm,ndxVal,fldNm):

    def GetRowFldValue(self,rowNum,fldNm):
        self.ConfirmFldExists(fldNm)
        if rowNum >= self.tblSz:
            RaiseException('ERROR: row number %d out of range.'%(rowNum))
        return self.tblDict[fldNm][rowNum]
    # def GetRowFldValue(self,rowNum,fldNm):

    def CopyFld(self,rsltFld,argFld):
        self.ConfirmFldExists(fldNm)
        self.tblDict[rsltFld] = self.tblDict[argFld1]
    # def CopyFld()

    def GetFld(self,fldNm):
        self.ConfirmFldExists(fldNm)
        return self.tblDict[fldNm]
    # def GetFld()

    def GetFldNms(self):
        return self.tblDict.keys()
    # def GetFldNms(self):

    def GetNdxNms(self):
        return self.ndxDict.keys()
    # def GetNdxNms(self):

    def GetTblSz(self):
        return self.tblSz
    # def GetNdxNms(self):

    def DeleteFld(self,fldNm):
        if self.FldExists(fldNm):
            del(tblDict[fldNm])
    # def DeleteFld()

    def DeleteNdx(self,fldNm):
        if self.FldExists(fldNm):
            del(tblDict[fldNm])
    # def DeleteFld()

    def WriteTbl(self,fName):
        outFile = open(fName,'wb')
        pickle.dump(self.tblDict,outFile)
        outFile.close()
    # def WriteTbl(self,fName):
            
    def ReadTbl(self,fName):
        inFile = open(fName,'rb')
        pickle.load(inFile)
        inFile.close()
    # def ReadTbl(self,fName):   
    
    # Operator methods

    def DifFlds(self,rsltFld,argFld1,argFld2):
        self.ConfirmFldsExist([argFld1,argFld2])

        if rsltFld not in self.tblDict.keys():
            self.AddFld(rsltFld,[0]*self.tblSz)
        else:
            self.tblDict[rsltFld] = [0]*self.tblSz

        for i in range(self.tblSz):
            if self.tblDict[rsltFld][i] != None and self.tblDict[argFld2][i] != None:
                self.tblDict[rsltFld][i] = self.tblDict[argFld1][i] - self.tblDict[argFld2][i]

        # for i in range(tblSz):
    # def DifFlds(rsltFld,argFld1,argFld2):

    def MinFlds(self,rsltFld,argFlds):
        self.ConfirmFldsExist(argFlds)
        if len(argFlds) < 1:
            RaiseException("Error: min requires one or more argument fields!")

        if rsltFld not in self.tblDict.keys():
            self.AddFld(rsltFld,[float('inf')]*self.tblSz)
        else:
            self.tblDict[rsltFld] = [float('inf')]*self.tblSz

        for i in range(self.tblSz):
            for key in argFlds:
                if self.tblDict[key][i] != None:
                    self.tblDict[rsltFld][i] = min(self.tblDict[rsltFld][i], self.tblDict[key][i])
        # for i in range(tblSz):
    # def MinFlds(rsltFld,argFlds):

    def MaxFlds(self,rsltFld,argFlds):
        self.ConfirmFldsExist(argFlds)
        if len(argFlds) < 1:
            RaiseException("Error: max requires one or more argument fields!")

        if rsltFld not in self.tblDict.keys():
            self.AddFld(rsltFld,[float('-inf')]*self.tblSz)
        else:
            self.tblDict[rsltFld] = [float('-inf')]*self.tblSz

        for i in range(self.tblSz):
            for key in argFlds:
                if self.tblDict[key][i] != None:
                    self.tblDict[rsltFld][i] = max(self.tblDict[rsltFld][i], self.tblDict[key][i])
        # for i in range(tblSz):
    # def MaxFlds(rsltFld,argFlds):

    def SumFlds(self,rsltFld,argFlds,argWts):
        self.ConfirmFldsExist(argFlds)
        if len(argFlds) < 2:
            RaiseException("Error: sum requires two or more addends")

        if len(argWts) != 0 and len(argWts) != len(argFlds):
            RaiseException("Error: wtdMean, number of fields must equal number of weights!")

        if len(argWts) == 0: # unweighted mean uses arg weights of 1
            argWts = [1] * len(argFlds)

        if rsltFld not in self.tblDict.keys():
            self.AddFld(rsltFld,[0]*self.tblSz)
        else:
            self.tblDict[rsltFld] = [0]*self.tblSz

        for i in range(self.tblSz):
            j = 0
            for key in argFlds:
                if self.tblDict[key][i] != None:
                    self.tblDict[rsltFld][i] = self.tblDict[rsltFld][i] + self.tblDict[key][i] * argWts[j]
                j = j+1
        # for i in range(tblSz):
    # def SumFlds(rsltFld,argFlds):

    def MeanFlds(self,rsltFld,argFlds,argWts):
        self.ConfirmFldsExist(argFlds)
        if len(argFlds) < 1:
            RaiseException("Error: wtdMean requires one or more argument fields!")
        if len(argWts) != 0 and len(argWts) != len(argFlds):
            RaiseException("Error: wtdMean, number of fields must equal number of weights!")

        if len(argWts) == 0: # unweighted mean uses arg weights of 1
            argWts = [1] * len(argFlds)

        if rsltFld not in self.tblDict.keys():
            self.AddFld(rsltFld,[0]*self.tblSz)
        else:
            self.tblDict[rsltFld] = [0]*self.tblSz

        for i in range(self.tblSz):
            opSum = 0
            divisor = 0
            j = 0
            for key in argFlds:
                if self.tblDict[key][i] != None:
                    opSum = opSum + self.tblDict[key][i] * argWts[j]
                    divisor = divisor + argWts[j]
                j = j+1

            if divisor == 0:
                self.tblDict[rsltFld][i] = None
            else:
                self.tblDict[rsltFld][i] = opSum / divisor
        # for i in range(tblSz):
    # def MeanFlds(rslt,argFldsAndWts):

    def LinearCvtFld(self,rsltFld,srcFld,x1,y1,x2,y2):
        self.ConfirmFldExists(srcFld)
        if x1 == x2 or y1 == y2:
            RaiseException("Error: linearCvt requires x1 != x2 and y1 != y2")
        if rsltFld not in self.tblDict.keys():
            self.AddFld(rsltFld,[None]*self.tblSz)
        else:
            self.tblDict[rsltFld] = [None]*self.tblSz

        # generate m and b for the linear equation:
        m = (y2 - y1) / (x2 - x1)
        b = -m * x1 + y1

        # apply conversion to srcFld
        for i in range(self.tblSz):
            if self.tblDict[srcFld][i] != None:
                self.tblDict[rsltFld][i] = m * self.tblDict[srcFld][i] + b
        # for i in range(self.tblSz):
    # def LinearCvt(self,rsltFld,srcFld,x1,y1,x2,y2):

    def ImposeMaxMinFld(self,rsltFld,srcFld,maxVal,minVal):
        self.ConfirmFldExists(srcFld)
        if maxVal < minVal:
            RaiseException("Error: imposeMaxMin requires maxVal >= minVal")
        if rsltFld not in self.tblDict.keys():
            self.AddFld(rsltFld,[None]*self.tblSz)

        # apply conversion to srcFld
        for i in range(self.tblSz):
            if self.tblDict[srcFld][i] != None:
                self.tblDict[rsltFld][i] = max(min(maxVal,self.tblDict[srcFld][i]),minVal)
    # def ImposeMaxMinFld(self,rsltFld,srcFld,maxVal,minVal):

    # Fuzzy logic operators (unless otherwise noted EMDS based)
    # Note: no checking done to see if values correspond to -1 to 0
    #   continuum
    
    def CvtToFuzzy(self,rsltFld,srcFld,falseThresh,trueThresh):
        self.ConfirmFldExists(srcFld)

        # if thresholds are default
        if falseThresh == MinForFuzzyLimit:
            falseThresh = self.MinOfFld(srcFld)
        elif falseThresh == MaxForFuzzyLimit:
            falseThresh = self.MaxOfFld(srcFld)

        if trueThresh == MinForFuzzyLimit:
            trueThresh = self.MinOfFld(srcFld)
        elif trueThresh == MaxForFuzzyLimit:
            trueThresh = self.MaxOfFld(srcFld)

        self.LinearCvtFld(rsltFld,srcFld,falseThresh,-1,trueThresh,1)
        self.ImposeMaxMinFld(rsltFld,rsltFld,1,-1)

        for i in range(self.tblSz):
            if self.tblDict[rsltFld][i] == None:
                self.tblDict[rsltFld][i] = NullFuzzyValue
        
    # def CvtToFuzzy(self,rsltFld,srcFld,falseThresh,trueThresh):

    def FuzzyNot(self,rsltFld,argFld):
        self.ConfirmFldExists(argFld)
        if rsltFld not in self.tblDict.keys():
            self.AddFld(rsltFld,[0]*self.tblSz)

        for i in range(self.tblSz):
            self.tblDict[rsltFld][i] = -self.tblDict[argFld1][i]

        # for i in range(tblSz):
    # def FuzzyNot()

    def FuzzyUnion(self,rsltFld,argFlds,argWts):
        self.MeanFlds(rsltFld,argFlds,argWts)
    # def FuzzyUnion(self,rsltFld,argFldsAndWts):

    def FuzzySelectedUnion(self,rsltFld,argFlds,trueOrFalse,numToConsider):
        self.ConfirmFldsExist(argFlds)
        if len(argFlds) < 1:
            RaiseException("Error: FuzzySelectedUnion requires one or more argument fields!")
        if trueOrFalse not in [-1,1]:
            RaiseException("Error: FuzzySelectedUnion variable trueOrFalse must be -1 or 1!")
        if numToConsider < 1:
            RaiseException("Error: FuzzySelectedUnion numToConsider must be <= 1!")

        if rsltFld not in self.tblDict.keys():
            self.AddFld(rsltFld,[0]*self.tblSz)
        else:
            self.tblDict[rsltFld] = [0]*self.tblSz

        for i in range(self.tblSz):
            values = []
            for key in argFlds:
                if self.tblDict[key][i] != None:
                    values.append(self.tblDict[key][i])

            if len(values) == 0:
                self.tblDict[rsltFld][i] = None
            else:
                # sort hi to lo or lo to hi and take desired # of fields
                if trueOrFalse == 1:
                    values = sorted(values,reverse=True)[0:numToConsider]
                else:
                    values = sorted(values)[0:numToConsider]

            # if len(values) == 0:
            self.tblDict[rsltFld][i] = sum(values) / len(values)
        # for i in range(self.tblSz):

    # def FuzzySelectedUnion(self,rsltFld,argFldsAndWts):

    def FuzzyAnd(self,rsltFld,argFlds,argWts):

        self.ConfirmFldsExist(argFlds)
        if len(argFlds) < 1:
            RaiseException("Error: fuzzyAnd requires one or more argument fields!")
        if len(argWts) != 0 and len(argWts) != len(argFlds):
            RaiseException("Error: fuzzyAnd, number of fields must equal number of weights!")
        
        if rsltFld not in self.tblDict.keys():
            self.AddFld(rsltFld,[0]*self.tblSz)

        if len(argWts) == 0:
            argWts = [1] * len(argFlds)

        divisor = sum(argWts)

        for i in range(self.tblSz):
            meanVal = 0
            minVal = float('inf')
            wtNdx = 0
            for key in argFlds:
                minVal = min(minVal,self.tblDict[key][i])
                meanVal = meanVal + self.tblDict[key][i] * argWts[wtNdx]
                wtNdx = wtNdx + 1

            meanVal = meanVal / divisor

            # for key in argFlds:
            self.tblDict[rsltFld][i] = minVal + (meanVal - minVal) * (minVal + 1) / 2
        # for i in range(self.tblSz):

    # def FuzzyAnd(self,rsltFld,argFldsAndWts):

    def FuzzyXOr(self,rsltFld,argFlds):
        self.ConfirmFldsExist(argFlds)
        if rsltFld not in self.tblDict.keys():
            self.AddFld(rsltFld,[0]*self.tblSz)
        if len(argFlds) < 1:
            RaiseException("Error: fuzzyXOr requires 1 or more fields to evaluate!")

        for i in range(self.tblSz):
            vals = []
            for key in argFlds:
                vals.append(self.tblDict[key][i])
            vals.sort()
            vals.reverse()

            if len(vals) == 1:
                rslt = vals[0]
            elif vals[0] == -1:
                rslt = -1
            elif vals[0] == 1 and vals[1] == 1:
                rslt = -1
            elif vals[0] == 1 and vals[1] == -1:
                rslt = 1
            else:
                rslt = (vals[0] - vals[1]) / (1 - -1)
            # if...elif...else
            self.tblDict[rsltFld][i] = rslt
            
        # for i in range(self.tblSz):
    # def FuzzyXOr(self,rsltFld,argFlds):

    def FuzzyOr(self,rsltFld,argFlds):
        self.MaxFlds(rsltFld,argFlds)
    # def FuzzyOr(self,rsltFld,argFlds):

    def FuzzyOrNeg(self,rsltFld,argFlds):
        self.MinFlds(rsltFld,argFlds)
    # def FuzzyOrNeg(self,rsltFld,argFlds):

    # Summary value from a single field
    def MaxOfFld(self,srcFld):
        self.ConfirmFldExists(srcFld)
        return max(self.tblDict[srcFld])
    # def MaxOfFld(self,srcFld):

    def MinOfFld(self,srcFld):
        self.ConfirmFldExists(srcFld)
        rtrn = float('inf')
        for val in self.tblDict[srcFld]:
            if val != None:
                rtrn = min(rtrn,val)
        return rtrn
    # def MinOfFld(self,srcFld):

    def SumOfFld(self,srcFld):
        self.ConfirmFldExists(srcFld)
        rtrn = 0
        for val in self.tblDict[srcFld]:
            if val != None:
                rtrn = rtrn + val
        return rtrn
    # def SumOfFld(self,srcFld):
        
    def MeanOfFld(self,srcFld):
        self.ConfirmFldExists(srcFld)
        rtrn = 0
        divisor = 0
        for val in self.tblDict[srcFld]:
            if val != None:
                rtrn = rtrn + val
                divisor = divisor + 1
        return rtrn / divisor
    # def MeanOfFld(self,srcFld):

    def PrintTbl(self):
        OutMsg("### Table ###")
        for fldName in self.tblDict.keys():
            OutMsg(fldName,':')
            OutMsg (self.tblDict[fldName])
    # def PrintTble(self):

    def PrintNdxs(self):
        OutMsg("### Indexes ###")
        for fldName in self.ndxDict.keys():
            OutMsg(fldName,':')
            OutMsg(self.ndxDict[fldName])
    # def PrintTble(self):

# class EEMSTbl:

class CmdTree:
    # A CmdTree consists of nodes representing a structured calculation diagram.
    # Each node consists of the name of the field created at the node and fields
    # necessary to create that field.
    # Each node is represented by a dictionary entry in fldNodes.
    # Each key in fldNodes is the name of the field calculated by that node
    # For each key, a second dictionary is used to store the parameters used
    #   to calculate the value of the calculated field.
    # fldNodes has the following structure:
    #
    # {calcFldName1:{'cmd':cmdName,'argFlds':[argumentFields],'argWts':[argWts]},
    # {calcFldName2:{'cmd':cmdName,'argFlds':[argumentFields],'argWts':[argWts]},
    #  . . .
    # {calcFldNamen:{'cmd':cmdName,'argFlds':[argumentFields],'argWts':[argWts]}}
    #
    # Depending on the cmd, argFlds, and/or argWts may or may not be present.
    # Valid cmdNames and the arguments required are as follows:
    #   cmd    argFlds  argWts  Description
    #   READ   none     none    Read a field from a field table
    #   OR     req      none    Fuzzy OR operation
    #   ORNEG  req      none    Fuzzy ORNEG operation
    #   XOR    req      none    Fuzzy XOR operation
    #   UNION  req      req     Fuzzy UNION operation
    #   AND    req      req     Fuzzy AND operation
    
    ### Need to add summary ops

    fldNodes = {} # dictionary of commands in tree indexed by result field name
    orderedNodes = [] # command keys (field names) in order of execution

    argSglFldCmds = ['COPYFIELD','NOT']
    argFldsOnlyCmds = ['OR','ORNEG','XOR','SUM','DIF','MIN','MAX','MEAN','UNION','AND']
    argFldsAndWtsCmds = ['WTDUNION','WTDAND','WTDMEAN','WTDSUM']
    specialCmds = ['READ','CVTTOFUZZY','SELECTEDUNION']

    # Commands with special checks:
    #   CVTTOFUZZY
    
    def HasNode(self,txtNode):
        tokens = txtNode.split(":")
        return self.fldNodes.has_key(tokens[0])
    # def HasNode(self,txtNode):

    def AddNode(self,txtNode):

        # txtNode is of the form:
        # FldName:Operator:ArgList1:ArgList2

        tokens = txtNode.split(":")
        tokens[1] = str.upper(tokens[1])
        
        if tokens[1] == 'READ':
            if len(tokens) != 2:
                RaiseException("ERROR: Command takes 0 arguments: %s"%(txtNode))
            
            self.fldNodes[tokens[0]] = {'cmd':tokens[1],'txt':txtNode}
            
        elif tokens[1] in self.argSglFldCmds:
            if len(tokens) != 3:
                RaiseException("ERROR: Command takes 1 arguments: %s"%(txtNode))
            
            argFldLst = tokens[2].split(',')
            if len(argFldLst) > 1:
                RaiseException("ERROR: Command takes 1 field arguments: %s"%(txtNode))
            
            self.fldNodes[tokens[0]] = {'cmd':tokens[1], 'argFlds':argFldLst,'txt':txtNode}

        elif tokens[1] in self.argFldsOnlyCmds:
            if len(tokens) != 3:
                RaiseException("ERROR: Command takes argFlds only: %s"%(txtNode))
            
            self.fldNodes[tokens[0]] = {'cmd':tokens[1],'argFlds':tokens[2].split(','),'txt':txtNode}
            
        elif tokens[1] == 'SELECTEDUNION':
            if len(tokens) != 4:
                RaiseException("ERROR: Command takes argFlds and numerics: %s"%(txtNode))
            if len(tokens[3].split(',')) != 2:          
                RaiseException("ERROR: Command uses exactly 2 numerics: %s"%(txtNode))

            # parse the command into its dictionary entry
            self.fldNodes[tokens[0]] = {}
            self.fldNodes[tokens[0]]['cmd'] = tokens[1]
            self.fldNodes[tokens[0]]['argFlds'] = tokens[2].split(',')
            self.fldNodes[tokens[0]]['trueOrFalse'] = int(tokens[3].split(',')[0])
            self.fldNodes[tokens[0]]['numToConsider'] = int(tokens[3].split(',')[1])
            self.fldNodes[tokens[0]]['txt'] = txtNode

        elif tokens[1] in self.argFldsAndWtsCmds:
            if len(tokens) != 4:
                RaiseException("ERROR: Command takes argFlds and weights: %s"%(txtNode))
            if len(tokens[2].split(',')) != len(tokens[3].split(',')):          
                RaiseException("ERROR: Command's argFlds and argWts must be same size: %s"%(txtNode))

            self.fldNodes[tokens[0]] = {'cmd':tokens[1],'argFlds':tokens[2].split(','),'argWts':tokens[3].split(','),'txt':txtNode}
            self.fldNodes[tokens[0]]['argWts'] = [float(item) for item in self.fldNodes[tokens[0]]['argWts']]

        elif tokens[1] == 'CVTTOFUZZY':
            if len(tokens) != 4:
                RaiseException("ERROR: Command takes 1 argFlds and 2 thresholds: %s"%(txtNode))
       
            self.fldNodes[tokens[0]] = {'cmd':tokens[1],'argFlds':[tokens[2]],'thresholds':tokens[3].split(','),'txt':txtNode}
            self.fldNodes[tokens[0]]['thresholds'] = [float(item) for item in self.fldNodes[tokens[0]]['thresholds']]

        else:
            RaiseException("ERROR: Illegal command: %s"%(tokens[1]))

        # if tokens[1] == "READ":...elif...else

    # def AddNode(self,txtNode):

    def OrderNodes(self):

        # orders the commands for execution. While doing so, detects
        # circular logic in commands and commands whose dependent
        # fields are undefined

        # check all commands to ensure that every dependent field is defined

        for fld in self.fldNodes.keys():
            if self.fldNodes[fld]['cmd'] not in ['READ']:
                for dependFld in self.fldNodes[fld]['argFlds']:
                    if dependFld not in self.fldNodes.keys():
                        RaiseException("ERROR: field %s depends on undefined field %s"%(fld,dependFld))
        # for fld in self.fldNodes.keys():

        # order the fields for execution

        # reinitialize ordered commands
        self.OrderedNodes = []
        nodesToBeOrdered = self.fldNodes.keys()

        while len(nodesToBeOrdered) > 0:
            
            startingNodesToBeOrderedLen = len(nodesToBeOrdered)

            # Step backwards through nodesToBeOrdered so that popping
            # elements does not interfere with indexing
            ndxs = range(len(nodesToBeOrdered))
            ndxs.reverse()
            for ndx in ndxs:
                
                fld = nodesToBeOrdered[ndx]

                if self.fldNodes[fld]['cmd'] == 'READ':
                    fldIsIndependent = True

                else:
                
                    # check if all the fields the current field depends on are
                    # in the orderedNodes list, i.e. have been found to be definable
                    
                    dependFlds = self.fldNodes[fld]['argFlds']
                    fldIsIndependent = True
                    
                    for dependFld in dependFlds:
                        if dependFld not in self.OrderedNodes:
                            fldIsIndependent = False
                            break
                        # if dependFld not in self.OrderedNodes:
                    # for dependFld in dependFlds:

                # if self.fldNodes[fld][0] == "READ"...else:...
                    
                if fldIsIndependent:
                    self.OrderedNodes.append(fld)
                    nodesToBeOrdered.pop(ndx)
                    break
                # if fldIsIndependent:
                
            # for ndx in ndxs:
            
            # if no independent fields were found, then there is
            # a dependency cycle in the structure of the commands
            
            if startingNodesToBeOrderedLen == len(nodesToBeOrdered):
                OutMsg('ERROR: Circular logic or missing dependency in this set of commands:')
                for fldNm in nodesToBeOrdered:
                    OutMsg(self.fldNodes[fldNm]['txt'])
                RaiseException("ERROR: Circular logic in field dependencies")

        # while len(nodesToBeOrdered) > 0:
    # def OrderNodes(self):
        
    def PrintOrderedNodes(self):

        OutMsg('Commands in order of Execution:')
        for fldNm in self.OrderedNodes:
            OutMsg('\t%s'%(self.fldNodes[fldNm]['txt']))

    # def PrintOrderedNodes(self):

    def GetModelImage(self):
        # order nodes to insure no cicular or missing dependencies
        self.OrderNodes()

        # find the top node(s)

        depends = []
        for key in self.fldNodes.keys():
            if self.fldNodes[key]['cmd'] != 'READ':
                depends = depends + self.fldNodes[key]['argFlds']

        topNodeFlds = []
        for key in self.fldNodes.keys():
            if key not in depends:
                topNodeFlds.append(key)

        # order nodes by dependency, tracking depth

        treeImage = []
        for topNodeFld in topNodeFlds:
            self.ParseDict(topNodeFld,treeImage,0)

        # return list

        return treeImage
    # def GetModelImage(self):

    def ParseDict(self,nodeFld,treeImage, lvl):
    # parses the dictionary into a dependency tree

        treeImage.append((nodeFld,lvl))
        lvl = lvl + 1

        if self.fldNodes[nodeFld]['cmd'] != 'READ':
            for subFld in self.fldNodes[nodeFld]['argFlds']:
                self.ParseDict(subFld, treeImage,lvl)

        return treeImage

    # def ParseDict(self,nodeFld,treeImage, lvl)
    
    def DisplayCmdTree(self):
        OutMsg('Entire Command Tree:\n')
        for fldTuple in self.GetModelImage():
            # Indent based on level
#            OutMsg('%s%s'%(fldTuple[1] * '\t', self.fldNodes[fldTuple[0]]['txt']))
            OutMsg('%s%s'%(fldTuple[1] * '  |', self.fldNodes[fldTuple[0]]['txt']))
        OutMsg('\n')

    # def DisplayCmdTree(self);

    def GetCmdNode(self,fld):
        return self.fldNodes[fld]
    # def GetCmdNode(self,fld):

    def GetNodeOrder(self):
        return self.OrderedNodes
    # def GetOrderedNodes(self):

    def GetNodeFldNms(self):
        return self.fldNodes.keys()
    # def GetNodeFldNms(self):

    def GetReadFldNms(self):
        rtrnLst = []
        for fldNm in self.fldNodes.keys():
            if self.fldNodes[fldNm]['cmd'] == 'READ':
                rtrnLst.append(fldNm)
        return rtrnLst
    # def GetReadFldNms(self):

    def IsReadFldNm(self,fldNm):
        rtrn = False
        if fldNm in self.fldNodes.keys():
            if self.fldNodes[fldNm]['cmd'] == 'READ':
                rtrn = True
        rtrn = True
    # def IsReadFldNm(self,fldNm):
            
    def GetComputeFldNms(self):
        rtrnLst = []
        for fldNm in self.fldNodes.keys():
            if self.fldNodes[fldNm]['cmd'] != 'READ':
                rtrnLst.append(fldNm)
        return rtrnLst
    # def GetComputeFldNms(self):

    def IsComputeFldNm(self,fldNm):
        rtrn = False
        if fldNm in self.fldNodes.keys():
            if self.fldNodes[fldNm]['cmd'] != 'READ':
                rtrn = True
        rtrn = True
    # def IsComputFldNm(self,fldNm):
            
    def GetNodeFldNms(self):
        return self.fldNodes.keys()
    # def GetOrderedNodes(self):

    def LoadCmdFile(self, fNm):
        inFile = open(fNm,'r')
        for line in inFile:
            line = line.rstrip()
            if not re.search('^#',line):
                self.AddNode(line)
        # for line in inFile:
        inFile.close()
        self.OrderNodes()
    # def GetNodesFromFile(self, fNm):

# class CmdTree:

class EEMSModel(EEMSTbl,CmdTree):

    def ExecModel(self):
        if DEBUG:
            OutMsg('Calculating Fields:')
        for fld in self.GetNodeOrder():
            if DEBUG:
                OutMsg('\t%s'%(fld))
            if DEBUG:
                OutMsg('\tExecuting **%s:%s**'%(fld,self.GetCmdNode(fld)))
            self.ExecNodeCmd(fld,self.GetCmdNode(fld))
        # for fld in orderedCmdNodeFlds:
    # def ExecModel(self):

    def ExecNodeCmd(self,rsltFldNm,cmdNode):

        if cmdNode['cmd'] == 'READ':
            self.ConfirmFldExists(rsltFldNm)
    
        elif cmdNode['cmd'] == 'COPYFIELD':
            self.CopyFld(rsltFldNm,cmdNode['argFlds'][0])

        elif cmdNode['cmd'] == 'CVTTOFUZZY':
            self.CvtToFuzzy(rsltFldNm,cmdNode['argFlds'][0],
                              cmdNode['thresholds'][0],cmdNode['thresholds'][1])

        elif cmdNode['cmd'] == 'NOT':
            self.FuzzyNot(rsltFldNm,cmdNode['argFlds'][0])

        elif cmdNode['cmd'] == 'OR':
            self.FuzzyOr(rsltFldNm,cmdNode['argFlds'])

        elif cmdNode['cmd'] == 'ORNEG':
            self.FuzzyOrNeg(rsltFldNm,cmdNode['argFlds'])

        elif cmdNode['cmd'] == 'XOR':
            self.FuzzyXOr(rsltFldNm,cmdNode['argFlds'])

        elif cmdNode['cmd'] == 'SUM':
            self.SumFlds(rsltFldNm,cmdNode['argFlds'],[])

        elif cmdNode['cmd'] == 'DIF':
            self.DifFlds(rsltFldNm,cmdNode['argFlds'][0],cmdNode['argFlds'][1])

        elif cmdNode['cmd'] == 'MIN':
            self.MinFlds(rsltFldNm,cmdNode['argFlds'])

        elif cmdNode['cmd'] == 'MAX':
            self.MaxFlds(rsltFldNm,cmdNode['argFlds'])

        elif cmdNode['cmd'] == 'MEAN':
            self.MeanFlds(rsltFldNm,cmdNode['argFlds'],[])

        elif cmdNode['cmd'] == 'UNION':
            self.FuzzyUnion(rsltFldNm,cmdNode['argFlds'],[])

        elif cmdNode['cmd'] == 'SELECTEDUNION':
            self.FuzzySelectedUnion(rsltFldNm,cmdNode['argFlds'],cmdNode['trueOrFalse'],cmdNode['numToConsider'])

        elif cmdNode['cmd'] == 'AND':
            self.FuzzyAnd(rsltFldNm,cmdNode['argFlds'],[])

        elif cmdNode['cmd'] == 'WTDUNION':
            self.FuzzyUnion(rsltFldNm,cmdNode['argFlds'],cmdNode['argWts'])

        elif cmdNode['cmd'] == 'WTDAND':
            self.FuzzyAnd(rsltFldNm,cmdNode['argFlds'],cmdNode['argWts'])

        elif cmdNode['cmd'] == 'WTDMEAN':
            self.MeanFlds(rsltFldNm,cmdNode['argFlds'],cmdNode['argWts'])

        elif cmdNode['cmd'] == 'WTDSUM':
            self.SumFlds(rsltFldNm,cmdNode['argFlds'],cmdNode['argWts'])

        else:
            RaiseException("ERROR: Invalid command: %s."%(cmdNode['cmd']))

    # def execNodeCmd(self.,rsltFldNm,cmdNode):
# class EEMSModel:

def PrintEEMSHdr():
    OutMsg('+------------------------------------------------------------------+')
    OutMsg('|                                                                  |')
    OutMsg('|         EEMS - Environmental Evaluation Modeling System          |')
    OutMsg('|                                                                  |')
    OutMsg('| Implementation for ArcMap Model Builder                          |')
    OutMsg('| Version: Alpha 0.0                                               |')
    OutMsg('|                                                                  |')
    OutMsg('+------------------------------------------------------------------+')
# def PrintEEMSHdr:

########################################################################
# Processing Starts HERE
########################################################################

PrintEEMSHdr()

# For the Model Logic Check tool
tmp = arcpy.GetParameterAsText(0)
if tmp == "CHECKLOGIC":
    cmdFileNm = arcpy.GetParameterAsText(1)
    myEEMSModel = EEMSModel()
    myEEMSModel.LoadCmdFile(cmdFileNm)
    myEEMSModel.DisplayCmdTree()
    exit(0)
# end if tmp

# Read input variables

modelTblNm = tmp
cmdFileNm = arcpy.GetParameterAsText(1)

OutMsg('modelTblNm:\n\t%s'%(modelTblNm))
OutMsg('cmdFileNm:\n\t%s'%(cmdFileNm))

myEEMSModel = EEMSModel()
myEEMSModel.LoadCmdFile(cmdFileNm)

myEEMSModel.DisplayCmdTree()

# check for presence of all read fields in input table

missingReadFlds = []
modelShapefileFldNms = []

for fld in arcpy.ListFields(modelTblNm):
    modelShapefileFldNms.append(fld.name)

for fldNm in myEEMSModel.GetReadFldNms():
    if fldNm not in modelShapefileFldNms:
        missingReadFlds.append(fldNm)
# for fldNm in myEEMSModel.GetReadFldNms()

if len(missingReadFlds) > 0:
    RaiseException('Error: Field(s) missing from input table:\n\t%s'%('\n\t'.join(missingReadFlds)))
# if len(missingReadFlds) > 0:

# Now put the fields into EEMSTbl
#   For each field, read data into a list
#     add list to myEEMSTbl
#
# NOTE: Looping over fields before rows makes it MUCH faster!

OutMsg('Reading in Fields:')

for fldNm in myEEMSModel.GetReadFldNms():
    OutMsg('\t%s'%(fldNm))
    inData = []
    rows = arcpy.SearchCursor(modelTblNm)
    for row in rows:
        inData.append(row.getValue(fldNm))
    myEEMSModel.AddFld(fldNm,inData)
# for fld in arcpy.ListFields(modelTblNm):

OutMsg('Table size:')
OutMsg('\t%d'%(myEEMSModel.GetTblSz()))

# Execute the Model
myEEMSModel.ExecModel()

OutMsg('Computing Field Values:')
for fldNm in myEEMSModel.GetComputeFldNms():
    OutMsg('\t%s'%(fldNm))
# Create list of fields that were created/updated
# by model

# Add the new columns to the shapefile

OutMsg('Adding Columns to Table:')

fldNms = myEEMSModel.GetComputeFldNms()
fldNms.sort()

for fldNm in fldNms:
    if fldNm not in arcpy.ListFields(modelTblNm):
        OutMsg('\t%s'%(fldNm))
        arcpy.AddField_management(modelTblNm, fldNm, "DOUBLE",10,4)
# for fldNm in myEEMSModel.GetComputeFldNms():

# Fill in the new values
        
OutMsg('Updating Fields in Table:')
for fldNm in fldNms:
    OutMsg('\t%s'%(fldNm))
    rows = arcpy.UpdateCursor(modelTblNm)
    rowNdx = 0
#    OutMsg('Value: %f'%(myEEMSTbl.getRowFldValue(rowNdx,fldNm)))

    for row in rows:
        value = myEEMSModel.GetRowFldValue(rowNdx,fldNm)
        # if rowNdx % 100 == 0:
        #     arcpy.AddMessage('\t\tValue: %f'%(value))
        row.setValue(fldNm,value)
        rowNdx = rowNdx + 1
        rows.updateRow(row)
    # for row in rows:

# for fldNm in myEEMSModel.GetComputeFldNms():

# Write EEMSTbl to arc table in gdb

# myEEMSModel.WriteFldsToCSV('F:\\Tehach\\TestModelData\\TehachTest.csv')

arcpy.SetParameter(2,modelTblNm)

# Profit

