##Summarize Raster by Reporting Unit
#################################################################################
############################################################################
############################################################################

#########
############
###############################################################################################
##############################
###########################
######################################################################################################

#########
#############################################################
###################
############################################################

import arcpy,os,traceback

try:
    arcpy.CheckOutExtension("spatial")
    arcpy.env.pyramid="NONE"
    arcpy.env.rasterStatistics="NONE"
    #arcpy.env.cartographicCoordinateSystem = "PROJCS['USA_Contiguous_Albers_Equal_Area_Conic_USGS_version',GEOGCS['GCS_North_American_1983',DATUM['D_North_American_1983',SPHEROID['GRS_1980',6378137.0,298.257222101]],PRIMEM['Greenwich',0.0],UNIT['Degree',0.0174532925199433]],PROJECTION['Albers'],PARAMETER['False_Easting',0.0],PARAMETER['False_Northing',0.0],PARAMETER['Central_Meridian',-96.0],PARAMETER['Standard_Parallel_1',29.5],PARAMETER['Standard_Parallel_2',45.5],PARAMETER['Latitude_Of_Origin',23.0],UNIT['Meter',1.0]]"

    inGrid=arcpy.GetParameterAsText(0)
    ruGrid=arcpy.GetParameterAsText(1)
    ruFC=arcpy.GetParameterAsText(2)
    statistic=arcpy.GetParameterAsText(3)
    outFC=arcpy.GetParameterAsText(4)

    #must be a geodatabase to contain the table we create
    if not arcpy.env.scratchWorkspace.count(".gdb"):
        arcpy.AddError("Environment -> Scratch Workspace must point at a geodatabase")
        raise ValueError("Environment -> Scratch Workspace must point at a geodatabase")

    zonalTable="%s/zonalTable"%(arcpy.env.scratchWorkspace)

    for item in [outFC,zonalTable]:
        if arcpy.Exists(item):
            arcpy.Delete_management(item)

    inputGrid=inGrid
    ruRaster=arcpy.Raster(ruGrid)
    inRaster=arcpy.Raster(inGrid)
    tmpGrid=None
    if not ruRaster.meanCellHeight==inRaster.meanCellHeight:
        arcpy.AddMessage("Resampling input grid to match reporting unit grid...")
        if not arcpy.env.scratchWorkspace:
            tmpWorkspace="c:/temp"
        elif arcpy.env.scratchWorkspace.count(".gdb"):
            tmpWorkspace=os.path.join(os.path.split(arcpy.env.scratchWorkspace)[0],"temp")
        if not os.path.exists(arcpy.env.scratchWorkspace):
            os.makedirs(tmpWorkspace)
        
        tmpGrid="%s/tmpResamp"%(tmpWorkspace)
        arcpy.Resample_management (inGrid, tmpGrid,ruRaster.meanCellHeight)
        inputGrid=tmpGrid
            
    arcpy.AddMessage("Creating zonal statistics...")
    
    #doesn't work with a geodatabase:
    #arcpy.sa.ZonalStatisticsAsTable (ruGrid, "Value", inputGrid, zonalTable)
    arcpy.sa.ZonalStatisticsAsTable (ruGrid, "Value", inputGrid, zonalTable, "DATA", statistic)
    
    rows = arcpy.SearchCursor(zonalTable)
    statisticAtt=statistic
    info=arcpy.Describe(zonalTable)
    for field in info.fields:
        if field.name==statisticAtt:
            attType=field.type
            if attType=="Single":
                attType="DOUBLE"
            precision=field.precision
            scale=field.scale
            break

    results=dict()    
    for row in rows: #only one row per zone; this is assumed to be 1:1 with summary unit FCs
        results[row.getValue("Value")]=row.getValue(statisticAtt)
    del row,rows

    arcpy.AddMessage("Adding results to output...")
    arcpy.CopyFeatures_management(ruFC,outFC)
    arcpy.AddField_management(outFC, statistic, attType, precision, scale, "","", "NULLABLE")
    rows=arcpy.UpdateCursor(outFC,"","","SHAPE,OBJECTID,%s"%(statistic))
    for row in rows:
        rowID=row.getValue("OBJECTID")
        value=0
        area=row.shape.area 
        if (results.has_key(rowID)) and area>0:
            value=results[rowID]
            row.setValue(statistic,value)
            rows.updateRow(row)
    del row,rows

    arcpy.Delete_management(zonalTable)
    if tmpGrid and arcpy.Exists(tmpGrid):
        arcpy.Delete_management(tmpGrid)
    
    arcpy.SetParameter(4,outFC)
    arcpy.SetParameter(5,statistic)

except:
    arcpy.AddError(traceback.format_exc())
    print traceback.format_exc()
    raise Exception("Error occurred during processing")



