#################### summarize_feature_class_by_reporting_units.py ######################
# Summarizes features within reporting units using particular statistic; uses area or length weighting for averaging polygons and lines respectively
# NOTE: reporting units that do not have any features are set to <Null>, NOT 0

# Inputs:
# Input Feature Class
# Reporting Unit Feature Class
# Output Feature Class  (this will be a copy of reporting unit feature class with summary field added)

#Outputs:
# Output feature class (so it can be consumed by later tools)
# Output field name
############################################################

import arcpy,os,traceback

try:
    workspace=arcpy.env.scratchWorkspace

    opNm=arcpy.GetParameterAsText(0) #in MIN,MAX,MEAN,SUM
    inFC=arcpy.GetParameterAsText(1)
    dissolveInFC=arcpy.GetParameter(2)
    inField=arcpy.GetParameterAsText(3)
    ruFC=arcpy.GetParameterAsText(4)
    outFC=arcpy.GetParameterAsText(5)

    disFC="%s/temp_dis"%(workspace)
    intFC="%s/temp"%(workspace)
    for fc in [outFC,disFC,intFC]:
        if arcpy.Exists(fc):
            arcpy.Delete_management(fc)
            
    #need to dissolve first or risk double counting overlapping features!
    inputFC=inFC
    if dissolveInFC:        
        arcpy.AddMessage("Dissolving...")
        arcpy.Dissolve_management(inFC,disFC,[inField],"","SINGLE_PART")
        inputFC=disFC            
    arcpy.AddMessage("Intersecting with reporting units...")
    arcpy.Intersect_analysis ([ruFC,inputFC],intFC)

    FID_att="FID_%s"%(os.path.split(arcpy.mapping.Layer(ruFC).dataSource)[1])
    arcpy.AddMessage("Adding results to output...")
    arcpy.CopyFeatures_management(ruFC,outFC)
    arcpy.AddField_management(outFC, opNm, "DOUBLE", 10, 4, "","", "NULLABLE")

    arcpy.AddMessage("Tallying results...")
    results=dict()
    featureType=None

    rows=arcpy.SearchCursor(intFC)
    if opNm=="MEAN":
        #do an area-weighted average of polygons, length-weighted average of lines, and straight average of points
        totals=dict()
        for row in rows:
            rowID=row.getValue(FID_att)
            if not featureType:
                featureType=row.shape.type
            if not totals.has_key(rowID):
                totals[rowID]=0
            if not results.has_key(rowID):
                results[rowID]=[]
            quantity=1.0
            if featureType.count("polygon"):
                quantity=row.shape.area
            elif featureType.count("line"):
                quantity=row.shape.length
            value=row.getValue(inField)
            if value != None:
                totals[rowID]+=quantity
                results[rowID].append([value,quantity])
        del row,rows

        rows=arcpy.UpdateCursor(outFC,"","","SHAPE,OBJECTID,%s"%(opNm))
        for row in rows:
            rowID=row.getValue("OBJECTID")
            value=0
            weight=0
            if results.has_key(rowID):
                value=sum([entry[0]*(entry[1]/totals[rowID]) for entry in results[rowID]])
                row.setValue(opNm,value)
                rows.updateRow(row)
        del row,rows

    else:
        #do a straight min/max/sum        
        for row in rows:
            rowID=row.getValue(FID_att)
            if not results.has_key(rowID):
                results[rowID]=[]
            results[rowID].append(row.getValue(inField))
        del row,rows

        rows=arcpy.UpdateCursor(outFC,"","","SHAPE,OBJECTID,%s"%(opNm))
        for row in rows:
            rowID=row.getValue("OBJECTID")
            value=0
            if results.has_key(rowID):
                if opNm=="MIN":
                    value=min(results[rowID])
                elif opNm=="MAX":
                    value=max(results[rowID])
                elif opNm=="SUM":
                    value=sum(results[rowID])
                row.setValue(opNm,value)
                rows.updateRow(row)
        del row,rows

    #delete intermediate
    for fc in [disFC,intFC]:
        if arcpy.Exists(fc):
            arcpy.Delete_management(fc)

    arcpy.SetParameter(5,outFC)
    arcpy.SetParameter(6,opNm)

except:
    arcpy.AddError(traceback.format_exc())

                                                       








